"""
Subgraph transforms for visualization, graph algorithms, etc.

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

import typing

from icecream import ic  #  type: ignore # pylint: disable=W0611
from tqdm import tqdm  # type: ignore
import pandas as pd  # type: ignore
import pyvis.network  # type: ignore
import networkx as nx  # type: ignore
import numpy as np  # type: ignore

from .kglab import KnowledgeGraph
from .topo import Measure
from .pkg_types import NodeLike, RDF_Node, RDF_Triple
from .algebra import AlgebraMixin
from .networks import NetAnalysisMixin
from .util import get_gpu_count


if get_gpu_count() > 0:
    import cudf  # type: ignore
    import cugraph # type: ignore # pylint: disable=W0611


class Subgraph:
    """
Base class for projection of an RDF graph into an *algebraic object* such as a *vector*,
 *matrix*, or *tensor* representation, to support integration with non-RDF graph libraries.
 In other words, this class provides means to vectorize selected portions of a graph as a
 [*dimension*](https://mathworld.wolfram.com/Dimension.html).
 See <https://derwen.ai/docs/kgl/concepts/#subgraph>

Features support several areas of use cases, including:

  * label encoding
  * vectorization (parallel processing)
  * graph algorithms
  * visualization
  * embedding (deep learning)
  * probabilistic graph inference (statistical relational learning)

The base case is where a *subset* of the nodes in the source RDF graph get represented as
a *vector*, in the `node_vector` member. This provides an efficient *index* on a constructed
*dimension*, solely for the context of a specific use case.
    """
    kg: KnowledgeGraph
    nx_graph: typing.Optional[nx.DiGraph] = None

    def __init__ (
        self,
        kg: KnowledgeGraph,
        *,
        preload: list = None,
        ) -> None:
        """
Constructor for creating and manipulating a *subgraph* as a [*vector*](https://mathworld.wolfram.com/Vector.html),
projecting from an RDF graph represented by a `KnowledgeGraph` object.

    kg:
the source RDF graph

    preload:
an optional, pre-determined list to pre-load for *label encoding*
        """
        self.kg = kg

        if preload:
            self.node_vector = preload
        else:
            self.node_vector = []


    def transform (
        self,
        node: NodeLike,
        ) -> int:
        """
Transforms a node in an RDF graph to an integer value, as a unique identifier with the closure of a specific use case.
The integer value can then be used to index into an *algebraic object* such as a *matrix* or *tensor*.
Effectvely, this method is similar to a [`sklearn.preprocessing.LabelEncoder`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html).

Notes:

  * the integer value is **not** a [*uuid*](https://tools.ietf.org/html/rfc4122) since it is only defined within the closure of a specific use case.
  * a special value `-1` represents the unique identifier for a non-existent (`None`) node, which is useful in data structures that have optional placeholders for links to RDF nodes

    node:
a node in the RDF graph

    returns:
a unique identifier (an integer index) for the `node` in the RDF graph
        """
        # NB: workaround to avoid Python calling `node.__nonzero__()`
        # https://amir.rachum.com/blog/2012/08/25/you-cant-handle-the-truth/

        # otherwise, if `node` is a literal (e.g., an int, float, string,
        # duration, etc.) that evals to any zero value then this function
        # would behave incorrectly

        if node is None:
            # null case
            return -1

        if not node in self.node_vector:
            self.node_vector.append(node)

        return self.node_vector.index(node)


    def inverse_transform (
        self,
        id: int,
        ) -> NodeLike:
        """
Inverse transform from an integer to a node in the RDF graph, using the identifier as an index into the node vector.

    id:
an integer index for the `node` in the RDF graph

    returns:
node in the RDF graph
        """
        if id < 0:
            return None

        return self.node_vector[id]


    def n3fy (
        self,
        node: RDF_Node,
        ) -> typing.Any:
        """
Wrapper for RDFlib [`n3()`](https://rdflib.readthedocs.io/en/stable/utilities.html?highlight=n3#serializing-a-single-term-to-n3) and [`toPython()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=toPython#rdflib.Variable.toPython) to serialize a node into a human-readable representation using N3 format.
This method provides a convenience, which in turn calls `KnowledgeGraph.n3fy()`

    node:
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node)

    returns:
text (or Python object) for the serialized node
        """
        return self.kg.n3fy(node)

    def check_attributes(self):
        """
Check if needed attributes are set.

        returns:
None
        """
        if self.kg is None:
            raise ValueError(
                """`Subgraph`'s `kg` should be initialized:
                `kglab.Subgraph(kg)`"""
            )

        # create an empty `nx.DiGraph` if none is present
        if self.nx_graph is None:
            # NOTE: find a way to pass `bipartite` if needed
            self.nx_graph = self.build_nx_graph(nx.DiGraph())  # pylint: disable=E1101


class SubgraphMatrix (Subgraph, AlgebraMixin, NetAnalysisMixin):  # pylint: disable=W0223
    """
Projection of a RDF graph to a [*matrix*](https://mathworld.wolfram.com/AdjacencyMatrix.html) representation.
Typical use cases include integration with non-RDF graph libraries for *graph algorithms*.

SPARQL query text needs to define a subgraph as: `subject -> object`.
    """
    _SRC_DST_MAP: typing.List[str] = ["subject", "object"]

    def __init__ (
        self,
        kg: KnowledgeGraph,
        sparql: str,
        *,
        bindings: dict = None,
        src_dst: typing.List[str] = None,
        ) -> None:
        """
Constructor for creating and manipulating a *subgraph* as a [*matrix*](https://mathworld.wolfram.com/AdjacencyMatrix.html),
projecting from an RDF graph represented by a `KnowledgeGraph` object.

    kg:
the source RDF graph

    sparql:
text for a SPARQL query that yields pairs to project into the *subgraph*; by default this expects the query to return
 bindings for `subject` and `object` nodes in the RDF graph

    bindings:
initial variable bindings

    src_dst:
an optional map to override the  `subject` and `object` bindings expected in the SPARQL query results; defaults to `None`
        """
        super().__init__(kg=kg)
        self.sparql = sparql
        self.bindings:typing.Optional[dict] = bindings

        if src_dst:
            self.src_dst: typing.List[str] = src_dst
        else:
            self.src_dst = self._SRC_DST_MAP


    def build_df (
        self,
        *,
        show_symbols: bool = False,
        ) -> pd.DataFrame:
        """
Factory pattern to populate a [`pandas.DataFrame`](https://pandas.pydata.org/docs/reference/frame.html) object,
using transforms in this subgraph.

Note: this method is primarily intended for [`cuGraph`](https://docs.rapids.ai/api/cugraph/stable/) support.
Loading via a `DataFrame` is required in lieu of using the `nx.add_node()` approach.
Therefore the support for representing *bipartite* graphs is still pending.

    show_symbols:
optionally, include the symbolic representation for each node; defaults to `False`

    returns:
the populated `DataFrame` object; uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled
        """
        col_names: typing.List[str] = [ "src", "dst", "src_sym", "dst_sym" ]

        if self.sparql is None and self.kg.use_gpus is True:
            raise ValueError("""To use GPUs is necessary to provide a SPARQL query to define a subgraph:
                            `kglab.SubgraphMatrix(kg, sparql)` or `SubgraphTensor(...)`""")
        row_iter = self.kg.query(self.sparql, bindings=self.bindings)

        if not show_symbols:
            col_names = col_names[:2]

            rows_list: typing.List[dict] = [
                {
                    col_names[0]: self.transform(row[self.src_dst[0]]),
                    col_names[1]: self.transform(row[self.src_dst[1]]),
                }
                for row in row_iter
                ]
        else:
            rows_list = [
                {
                    col_names[0]: self.transform(row[self.src_dst[0]]),
                    col_names[1]: self.transform(row[self.src_dst[1]]),
                    col_names[2]: self.n3fy(row[self.src_dst[0]]),
                    col_names[3]: self.n3fy(row[self.src_dst[1]]),
                }
                for row in row_iter
                ]

        if self.kg.use_gpus is True:
            df = cudf.DataFrame(rows_list, columns=col_names)
        else:
            df = pd.DataFrame(rows_list, columns=col_names)

        return df


    def build_nx_graph (
        self,
        nx_graph: nx.DiGraph,
        *,
        bipartite: bool = False,
        ) -> nx.DiGraph:
        """
Factory pattern to populate a [`networkx.DiGraph`](https://networkx.org/documentation/latest/reference/classes/digraph.html) object, using transforms in this subgraph.
See <https://networkx.org/>

    nx_graph:
pass in an unpopulated [`networkx.DiGraph`](https://networkx.org/documentation/latest/reference/classes/digraph.html) object; must be a [`cugraph.DiGraph`](https://docs.rapids.ai/api/cugraph/stable/api.html#digraph) if GPUs are enabled

    bipartite:
flag for whether the `(subject, object)` pairs should be partitioned into *bipartite sets*, in other words whether the *adjacency matrix* is symmetric; ignored if GPUs are enabled

    returns:
the populated `NetworkX` graph object; uses the [RAPIDS `cuGraph` library](https://docs.rapids.ai/api/cugraph/stable/) if GPUs are enabled
        """
        if self.kg.use_gpus is True:
            df = self.build_df()
            nx_graph.from_cudf_edgelist(df, source="src", destination="dst")
        else:
            for row in self.kg.query(self.sparql, bindings=self.bindings):
                s_id = self.transform(row[self.src_dst[0]])
                s_label = self.n3fy(row[self.src_dst[0]])

                o_id = self.transform(row[self.src_dst[1]])
                o_label = self.n3fy(row[self.src_dst[1]])

                if bipartite:
                    nx_graph.add_node(s_id, label=s_label, bipartite=0)
                    nx_graph.add_node(o_id, label=o_label, bipartite=1)
                else:
                    nx_graph.add_node(s_id, label=s_label)
                    nx_graph.add_node(o_id, label=o_label)

                nx_graph.add_edge(s_id, o_id)

        return nx_graph


    def build_ig_graph (
        self,
        ig_graph: typing.Any,
        ) -> typing.Any:
        """
Factory pattern to populate an [`igraph.Graph`](https://igraph.org/python/doc/igraph.Graph-class.html) object, using transforms in this subgraph.
See <https://igraph.org/python/doc/>

Note that `iGraph` is somewhat notorious for being quite difficult to install correctly across a wide range of different platforms and environments.
Consequently this has been removed from being a dependency for `kglab`; to use `iGraph` please install and import it separately.

    ig_graph:
pass in an unpopulated [`igraph.Graph`](https://igraph.org/python/doc/igraph.Graph-class.html) object

    returns:
the populated  `iGraph` graph object
        """
        measure = Measure()
        measure.measure_graph(self.kg)
        keyset = measure.get_keyset(incl_pred=False)

        ig_graph.add_vertices(n=keyset)

        for row in self.kg.query(self.sparql, bindings=self.bindings):
            s_id = self.transform(row[self.src_dst[0]])
            o_id = self.transform(row[self.src_dst[1]])
            ig_graph.add_edges([ (s_id, o_id,) ])

        ig_graph.vs["label"] = ig_graph.vs["name"] # pylint: disable=E1136,E1137
        return ig_graph

    def _get_n_nodes(self):
        """ Return number of nodes counted from the adjancency matrix"""
        return self.to_adjacency().shape[0]

    def _get_n_edges(self):
        """ Return number of edges counted from the adjancency matrix"""
        return int(np.sum(self.to_adjacency()))


class SubgraphTensor (Subgraph):
    """
Projection of a RDF graph to a [*tensor*](https://mathworld.wolfram.com/Tensor.html) representation.
Typical use cases include integration with non-RDF graph libraries for *visualization* and *embedding*.
    """

    def __init__ (
        self,
        kg: KnowledgeGraph,
        *,
        excludes: list = None,
        ) -> None:
        """
Constructor for creating and manipulating a *subgraph* as a [*tensor*](https://mathworld.wolfram.com/Tensor.html),
projecting from an RDF graph represented by a `KnowledgeGraph` object.

    kg:
the source RDF graph

    excludes:
a list of RDF predicates to exclude from projection into the *subgraph*
        """
        super().__init__(kg=kg)

        if excludes:
            self.excludes = excludes
        else:
            self.excludes = []


    def as_tuples (
        self
        ) -> typing.Generator[RDF_Triple, None, None]:
        """
Iterator for enumerating the RDF triples to be included in the subgraph, used in factory patterns for visualizations.
This allows a kind of *lazy evaluation*.

    yields:
the RDF triples within the subgraph
        """
        for s, p, o in self.kg.rdf_graph():
            if not p in self.excludes:
                yield s, p, o


    def as_tensor_edges (
        self
        ) -> typing.Generator[typing.List[int], None, None]:
        """
Iterator for enumerating the edges connecting to each predicate in the
subgraph, to be used to represent the KG in `PyTorch`.

    yields:
a subject and object edge for each predicate, in tensor representation
        """
        for s, p, o in self.as_tuples():
            s_label = self.n3fy(s)
            s_id = self.transform(s_label)

            p_label = self.n3fy(p)
            p_id = self.transform(p_label)

            o_label = self.n3fy(o)
            o_id = self.transform(o_label)

            yield [s_id, o_id, 2 * p_id]
            yield [o_id, s_id, 2 * p_id + 1]


    def as_tensor (
        self,
        *,
        quiet: bool = True,
        ) -> typing.List[typing.Tuple[int, int, int]]:
        """
Represents the KG as an edge list where each predicate has edges
connecting to its subject and object.

This can be used to load a [`Tensor`](https://pytorch.org/docs/stable/tensors.html)
in PyTorch, for example:

```
edge_list = kg.as_tensor()
tensor = torch.tensor(edge_list, dtype=torch.long).t().contiguous()
```

    quiet:
boolean flag to disable `tqdm` progress bar calculation and output

    returns:
an edge list for the loaded tensor object
        """
        edge_list: list = list(
            edge_tuple
            for edge_tuple in tqdm(self.as_tensor_edges(), disable=quiet)  # pylint: disable=R1721
        )

        return edge_list


    ######################################################################
    ## visualization
    ##
    ## Automated Network Graph: The triples describing relationships
    ## between entities can be ingested into graph visualization tools
    ## to extend or create an analyst's account-specific network
    ## model.

    def pyvis_style_node (  # pylint: disable=R0201
        self,
        pyvis_graph: pyvis.network.Network,
        node_id: int,
        label: str,
        *,
        style: dict = None,
        ) -> None : # pylint: disable=R0201
        """
Adds a node into a [PyVis](https://pyvis.readthedocs.io/) network, optionally with styling info.

    pyvis_graph:
the [`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html?highlight=network#pyvis.network.Network) being used for *interactive visualization*

    node_id:
unique identifier for a node in the RDF graph

    label:
text label for the node

    style:
optional style dictionary
        """
        if not style:
            style = {}

        prefix = label.split(":")[0]

        if prefix in style:
            pyvis_graph.add_node(
                node_id,
                label=label,
                title=label,
                color=style[prefix]["color"],
                size=style[prefix]["size"],
            )
        else:
            pyvis_graph.add_node(
                node_id,
                label=label,
                title=label,
            )


    def build_pyvis_graph (
        self,
        *,
        notebook: bool = False,
        style: dict = None,
        ) -> pyvis.network.Network:
        """
Factory pattern to create a [`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html?highlight=network#pyvis.network.Network) object, populated by transforms in this subgraph.
See <https://pyvis.readthedocs.io/>

    notebook:
flag for whether or not the interactive visualization will be generated within a notebook

    style:
optional style dictionary

    returns:
a `PyVis` network object
        """
        pyvis_graph = pyvis.network.Network(notebook=notebook)

        if not style:
            style = {}

        for s, p, o in self.as_tuples():
            # label the subject
            s_label = self.n3fy(s)
            s_id = self.transform(s_label)
            self.pyvis_style_node(pyvis_graph, s_id, s_label, style=style)

            # label the object
            o_label = str(self.n3fy(o))
            o_id = self.transform(o_label)
            self.pyvis_style_node(pyvis_graph, o_id, o_label, style=style)

            # label the predicate
            p_label = self.n3fy(p)
            pyvis_graph.add_edge(s_id, o_id, label=p_label)

        return pyvis_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Queryable abstraction layer.
"""

from abc import ABC, abstractmethod
import typing

import pandas as pd  # type: ignore  # pylint: disable=E0401


class Queryable (ABC):
    """
Abstract class for query support.
    """

    @abstractmethod
    def query (
        self,
        query: str,
        ) -> typing.Iterable:
        """
Abstract method for querying.
        """
        pass  # pylint: disable=W0107


    @abstractmethod
    def query_as_df (
        self,
        query: str,
        ) -> pd.DataFrame:
        """
Abstract method for querying, which returns a pandas DataFrame.
        """
        pass  # pylint: disable=W0107



"""
Working with `SubgraphMatrix` as vectorized representation.
Additions to functionalities present in `subg.py`.
Integrate `scipy` and `scikit-learn` functionalities.

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

import typing

import networkx as nx

from .util import Mixin


class AlgebraMixin (Mixin):
    """
Provides methods to work with graph algebra using `SubgraphMatrix` data.

NOTE: provide optional Oxigraph support for fast in-memory computation
    """
    nx_graph: typing.Optional[nx.DiGraph] = None


    def to_undirected (self):
        """
Return the undirected adjancency matrix of the directed graph.

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        return nx.to_numpy_array(self.nx_graph.to_undirected())


    def to_adjacency (self):
        """
Return adjacency (dense) matrix for the KG.
[Relevant NetworkX interface](https://networkx.org/documentation/stable/reference/convert.html#id2)

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.to_numpy_array(self.nx_graph)


    def to_incidence (self):
        """
Return incidence (dense) matrix for the KG.
[Relevant scipy docs](https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.html)

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.incidence_matrix(self.nx_graph).toarray()


    def to_laplacian (self):
        """
Return Laplacian matrix for the KG. Graph is turned into undirected.
[docs](https://networkx.org/documentation/stable/reference/generated/networkx.linalg.laplacianmatrix.laplacian_matrix.html).
Lapliacian is also known as vertices degrees matrix.

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.laplacian_matrix(self.nx_graph.to_undirected()).toarray()


    def to_scipy_sparse (self):
        """
Return graph in CSR format (optimized for matrix-matrix operations).

        returns:
SciPy sparse matrix: Graph adjacency matrix.
        """
        self.check_attributes()
        return nx.to_scipy_sparse_array(self.nx_graph)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright
# mypy: ignore-errors
# pylint: disable-all

"""
WIP: Evolunationary Shape Prediction
"""

from collections import deque
import random
import typing

import pandas as pd  # type: ignore
import rdflib  # type: ignore

from .kglab import KnowledgeGraph
from .topo import Measure
from .subg import Subgraph
from .pkg_types import RDF_Node, SPARQL_Bindings
from .util import calc_quantile_bins, root_mean_square, stripe_column


SerializedEvoEdge = typing.Tuple[int, int]
SerializedEvoNode = typing.Tuple[int, typing.List[SerializedEvoEdge]]

Evolike = typing.TypeVar("Evolike", int, SerializedEvoNode)
SerializedEvoShape = typing.List[Evolike]

EvoShapeBoard = typing.List[SerializedEvoShape]
EvoShapeDistance = typing.Tuple[int, int, float]


class EvoShapeNode:
    """
    """

    def __init__ (
        self,
        *,
        uri: str = None,
        terminal: bool = False,
        ) -> None:
        """
        """
        self.uri = uri
        self.terminal = terminal
        self.done = terminal # initially
        self.edges: typing.List[EvoShapeEdge] = []


    def serialize (
        self,
        subgraph: Subgraph,
        ) -> SerializedEvoNode:
        """
        """
        edge_list = sorted([ (subgraph.transform(e.pred), subgraph.transform(e.obj.uri),) for e in self.edges ])
        return subgraph.transform(self.uri), edge_list


    @classmethod
    def deserialize (
        cls,
        dat: SerializedEvoNode,
        subgraph: Subgraph,
        uri_map: dict,
        *,
        root_node: "EvoShapeNode" = None,
        ) -> "EvoShapeNode":
        """
        """
        node_id, edge_list = dat
        uri = subgraph.inverse_transform(node_id)

        if uri in uri_map:
            node = uri_map[uri]
        else:
            node = EvoShapeNode(uri=uri)
            uri_map[uri] = node

        for p, o in edge_list:
            uri = subgraph.inverse_transform(o)

            if not uri and root_node:
                node = root_node
            elif uri in uri_map:
                obj = uri_map[uri]
            else:
                obj = EvoShapeNode(uri=uri)
                uri_map[uri] = obj

            edge = EvoShapeEdge(pred=str(subgraph.inverse_transform(p)), obj=obj)
            node.edges.append(edge)

        return node


class EvoShapeEdge (typing.NamedTuple):
    """
    """

    pred: str
    obj: EvoShapeNode


class EvoShape:
    """
    """

    def __init__ (
        self,
        kg: KnowledgeGraph,
        measure: Measure,
        ) -> None:
        """
        """
        self.kg = kg
        self.measure = measure
        self.root = EvoShapeNode(uri=None)
        self.nodes = set([self.root])


    def add_link (
        self,
        s: RDF_Node,
        p: RDF_Node,
        o: RDF_Node,
        ) -> None:
        """
        """
        edge = EvoShapeEdge(pred=p, obj=o)
        s.edges.append(edge)
        self.nodes.add(o)


    def serialize (
        self,
        subgraph: Subgraph,
        ) -> SerializedEvoShape:
        """
Transform to ordinal format which can be serialized/deserialized, using a `Subgraph` with a consistent ordering across different distributed processes.
        """
        d: deque = deque(sorted([ n.serialize(subgraph) for n in self.nodes.difference({self.root}) ]))
        d.appendleft(self.root.serialize(subgraph))
        d.appendleft(self.get_cardinality())
        return list(d)


    def deserialize (
        self,
        dat_list: SerializedEvoShape,
        subgraph: Subgraph,
        ) -> dict:
        """
Replace shape definition with parsed content.
        """
        dat_list.pop(0) # ignore the instances
        uri_map: dict = {}
        self.nodes = set()

        root_dat = dat_list.pop(0)
        self.root = EvoShapeNode.deserialize(root_dat, subgraph, uri_map)
        self.nodes.add(self.root)

        for node_dat in dat_list:
            node = EvoShapeNode.deserialize(node_dat, subgraph, uri_map, root_node=self.root)
            self.nodes.add(node)

        return uri_map


    def get_rdf (
        self
        ) -> typing.List[str]:
        """
For debugging purposes only:
since by definition, a shape is not fully qualified RDF.
        """
        rdf_list: typing.List[str] = []

        for node in list(self.nodes):
            if not node.uri:
                b = rdflib.BNode()
                subj = str(b.n3())
            else:
                subj = rdflib.term.URIRef(node.uri).n3(self.kg.rdf_graph().namespace_manager)

            for edge in node.edges:
                pred = rdflib.term.URIRef(edge.pred).n3(self.kg.rdf_graph().namespace_manager)
                obj = rdflib.term.URIRef(edge.obj.uri).n3(self.kg.rdf_graph().namespace_manager)
                triple = "{} {} {} .".format(subj, pred, obj)
                rdf_list.append(triple)

        return rdf_list


    def get_sparql (
        self
        ) -> SPARQL_Bindings:
        """
        """
        var_list: typing.List[str] = []
        clauses: typing.List[str] = []
        uri_map: dict = {}
        bindings: dict = {}
        node_list = list(self.nodes)

        for node_num, node in enumerate(node_list):
            # name the subject
            if not node.uri:
                subj = "?v{}".format(node_num)
                var_list.append(subj)
            elif node.uri in uri_map:
                subj = uri_map[node.uri]
            else:
                subj = "?node{}".format(node_num)
                uri_map[node.uri] = subj
                bindings[subj] = rdflib.URIRef(node.uri)

            # generate a clause for each tuple
            for edge_num in range(len(node.edges)):
                edge = node.edges[edge_num]

                if edge.pred in uri_map:
                    pred = uri_map[edge.pred]
                else:
                    pred = "?pred{}_{}".format(node_num, edge_num)
                    uri_map[edge.pred] = pred
                    bindings[pred] = rdflib.URIRef(edge.pred)

                if edge.obj.uri in uri_map:
                    obj = uri_map[edge.obj.uri]
                else:
                    obj = "?obj{}_{}".format(node_num, edge_num)
                    uri_map[edge.obj.uri] = obj
                    bindings[obj] = rdflib.URIRef(edge.obj.uri)

                clauses.append(" ".join([ subj, pred, obj ]))

        sparql = "SELECT DISTINCT {} WHERE {{ {} }}".format(" ".join(var_list), " . ".join(clauses))
        return sparql, bindings


    def get_cardinality (
        self
        ) -> int:
        """
        """
        sparql, bindings = self.get_sparql()
        return len(list(self.kg.query(sparql, bindings=bindings)))


    def calc_distance (
        self,
        other: "EvoShape",
        ) -> float:
        """
        """
        n0 = { n.uri for n in self.nodes }
        n1 = { n.uri for n in other.nodes }
        distance = len(n0.intersection(n1)) / float(max(len(n0), len(n1)))
        return distance


class ShapeFactory:
    """
    """

    def __init__ (
        self,
        kg: KnowledgeGraph,
        measure: Measure,
        ) -> None:
        """
        """
        self.kg = kg
        self.measure = measure
        self.subgraph = Subgraph(kg, preload=measure.get_keyset())

        # enum action space of possible RDF types (i.e., "superclasses")
        type_sparql = "SELECT DISTINCT ?n WHERE {[] rdf:type ?n}"
        self.type_list = [ r.n for r in kg.query(type_sparql) ]


    def new_shape (
        self,
        *,
        type_uri: str = None,
        ) -> EvoShape:
        """
        """
        es = EvoShape(self.kg, self.measure)

        if not type_uri:
            ## RANDOM CHOICE => RL observation?
            ## TODO: generate from gamma dist -- or specify
            type_uri = random.choice(self.type_list)  # nosec

        type_node = EvoShapeNode(uri=type_uri, terminal=True)
        es.add_link(es.root, rdflib.RDF.type, type_node)
        return es


class Leaderboard:
    """
    """

    _COLUMNS: typing.List[str] = ["instances", "nodes", "distance", "rank", "shape"]

    def __init__ (
        self
        ) -> None:
        """
        """
        self.df = pd.DataFrame([], columns=self._COLUMNS)


    def get_board (
        self
        ) -> EvoShapeBoard:
        """
        Return a list of shapes, i.e., dataframe without any metrics.
        """
        return list(self.df["shape"].to_numpy())


    @classmethod
    def compare (
        cls,
        shape: SerializedEvoShape,
        board: EvoShapeBoard,
        ) -> EvoShapeDistance:
        """
        Compare distances among the shapes.
        """
        n0 = { n for n, e in shape[1:] }

        if len(board) < 2:
            min_dist = 0.0
        else:
            distances: list = []

            for b in board:
                n1 = { n for n, e in b[1:] }
                d = len(n0.intersection(n1)) / float(max(len(n0), len(n1)))

                if d < 1.0:
                    distances.append(d)

            min_dist = min(distances)

        return int(shape[0]), len(n0), min_dist


    @classmethod
    def insert (
        cls,
        shape: SerializedEvoShape,
        board: EvoShapeBoard,
        ) -> pd.DataFrame:
        """
Rank this `shape` within a dataframe generated from the given `board` object.
        """
        board.append(shape)
        df1 = pd.DataFrame([ cls.compare(s, board) for s in board ], columns=cls._COLUMNS[:3])

        # normalize by column
        df2 = df1.apply(lambda x: x/x.max(), axis=0)
        bins = calc_quantile_bins(len(df2.index))

        # stripe each column to approximate a pareto front
        stripes = [ stripe_column(values, bins) for _, values in df2.items() ]
        df3 = pd.DataFrame(stripes).T

        # rank based on RMS of striped indices per row
        df1["rank"] = df3.apply(root_mean_square, axis=1)
        df1["shape"] = pd.Series(board, index=df1.index)

        # sort descending
        return df1.sort_values(by=["rank"], ascending=False)


    def get_position (
        self,
        shape: SerializedEvoShape,
        ) -> int:
        """
Calculate a *distance-from-bottom* metric for the given `shape`.
        """
        return len(self.df.index) - list(self.df["shape"].to_numpy()).index(shape) - 1


    def add_shape (
        self,
        shape: SerializedEvoShape,
        ) -> int:
        """
Insert the given `shape` into the leaderboard, returning its position metric.
        """
        self.df = self.insert(shape, self.get_board())
        return self.get_position(shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Visualization of SPARQL queries.
"""

import collections
import copy
import typing

import pyvis  # type: ignore # pylint: disable=E0401
import rdflib.paths  # type: ignore # pylint: disable=E0401
import rdflib.plugins.sparql  # type: ignore # pylint: disable=E0401
import rdflib.term  # type: ignore # pylint: disable=E0401


class GPViz:
    """
Class used to Visualize the graph pattern of a SPARQL query.
This source comes from <https://github.com/pebbie/sparqlgpviz>
modified to limit its dependencies to RDFlib and PyVis.

by Peb Ruswono Aryan <https://github.com/pebbie>
    """
    _IGNORE_KEYS = [
        "PV",
        "var",
        "_vars",
        "datasetClause",
        "expr",
        "op",
        "A",
        "lazy",
        "service_string",
        "term",
        "start",  # used by LIMIT
        "length"  # used by LIMIT
    ]

    # colorcet glasbey_category10 first 16 colors
    # https://colorcet.holoviz.org/user_guide/Categorical.html
    _DEFAULT_COLORS = [
        "#1f77b3",
        "#ff7e0e",
        "#2ba02b",
        "#d62628",
        "#9367bc",
        "#8c564b",
        "#e277c1",
        "#7e7e7e",
        "#bcbc21",
        "#16bdcf",
        "#3a0182",
        "#004201",
        "#0fffa8",
        "#5d003f",
        "#bcbcff",
        "#d8afa1",
    ]


    def __init__ (
        self,
        sparql: str,
        namespaces: typing.Dict[str, str],
        ) -> None:
        """
Constructor for GPViz, to visualize the given SPARQL query as a [`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html#pyvis.network.Network)

    sparql:
input SPARQL query to be visualized

    namespaces:
the namespaces for the corresponding RDF graph
        """
        self.namespaces: typing.Dict[str, str] = copy.deepcopy(namespaces)
        pq = rdflib.plugins.sparql.prepareQuery(sparql, initNs=self.namespaces)

        for prefix, uri in pq.prologue.namespace_manager.namespaces():
            if prefix not in self.namespaces:
                self.namespaces[prefix] = str(uri)

        self.blank_nodes: typing.List[str] = []
        self.values: typing.Dict[str, list] = collections.defaultdict(list)
        self.triples: list = self._find_triples(pq.algebra)


    def _set_values (
        self,
        algebra,
        ) -> None:
        """
Semiprivate helper function to set the labels for literal values in the SPARQL query.

    algebra:
the parsed [SPARQL Algebra Expression](https://rdflib.readthedocs.io/en/stable/_modules/rdflib/plugins/sparql/algebra.html#translateQuery)
        """
        for obj in algebra:
            k = list(obj.keys())[0]
            v = obj[k]
            lname = str(k)
            self.values[lname].append(v)


    def _find_triples_node (
        self,
        algebra: typing.Any,
        key: str,
        result: list,
        ) -> None:
        """
Semiprivate helper function to expand one node in the given SPARQL Algebra Expression (at the specified key) to accumulate more triples to visualize.

    algebra:
the parsed [SPARQL Algebra Expression](https://rdflib.readthedocs.io/en/stable/_modules/rdflib/plugins/sparql/algebra.html#translateQuery)

    key:
expression key to expand

    result:
list into which the triples get accumulated
        """
        if key not in self._IGNORE_KEYS:
            if key == "res" and isinstance(algebra[key], list):
                self._set_values(algebra[key])
            elif key == "value":
                for var_item in algebra:
                    var = var_item["var"]
                    self.values[str(var)] = []

                    value = var_item["value"]

                    if isinstance(value, list):
                        for var in algebra["var"]:
                            tmpval = value.pop(0)
                            self.values[str(var)].append(tmpval)
                    else:
                        self.values[str(var)].append(value)
            elif key == "triples":
                result.extend([ algebra.triples ])
            else:
                result.extend(self._find_triples(algebra[key]))


    def _find_triples (
        self,
        algebra: typing.Any,
        ) -> list:
        """
Semiprivate helper function to expand the SPARQL Algebra Expression to accumulate more triples to visualize.

    algebra:
the parsed [SPARQL Algebra Expression](https://rdflib.readthedocs.io/en/stable/_modules/rdflib/plugins/sparql/algebra.html#translateQuery)

    returns:
a list of triples to visualize
        """
        result: list = []

        if isinstance(algebra, list):
            akg = algebra
        else:
            akg = [algebra]

        for algebra_node in akg:
            for key in dict(algebra_node).keys():
                self._find_triples_node(algebra_node, key, result)

        return result


    def _get_prefix (
        self,
        uri: rdflib.term.URIRef,
        ) -> typing.Optional[str]:
        """
Semiprivate helper function to determine the _prefix_ for the given namespace URI.

    uri:
namespace URI

    returns:
the prefix bound to this namespace URI
        """
        for k, v in sorted(self.namespaces.items(), key=lambda x: len(x[1]), reverse=True):
            if uri.startswith(str(v)):
                return k

        return None


    def _get_local_name (
        self,
        uri: rdflib.term.URIRef,
        ) -> str:
        """
Semiprivate helper function to localize the full URI for a node, if possible using its prefix.

    uri:
namespace URI

    returns:
localized name
        """
        prefix = self._get_prefix(uri)

        if prefix is not None:
            return prefix + ":" + str(uri).replace(self.namespaces[prefix], "")

        return str(uri)


    def _get_label (
        self,
        term: rdflib.term.Identifier,
        ) -> str:
        """
Semiprivate helper function to generate a localized label for the given RDF term.

    term:
the RDF term to be labeled

    returns:
generated label for the term
        """
        tname = str(term)

        if isinstance(term, rdflib.term.Variable):
            tname = "?" + tname
        elif isinstance(term, rdflib.term.URIRef):
            tname = self._get_local_name(term)
        elif isinstance(term, rdflib.term.BNode):
            if tname not in self.blank_nodes:
                self.blank_nodes.append(tname)

            tname = "_:bn" + str(self.blank_nodes.index(tname) + 1)
        elif isinstance(term, rdflib.paths.Path):
            if hasattr(term, "arg"):
                aname = self._get_local_name(str(term.arg))
                tname = tname.replace(str(term.arg), aname)
            elif hasattr(term, "args"):
                for arg in term.args:
                    tname = tname.replace(str(arg), self._get_local_name(arg))
            elif hasattr(term, "path"):
                aname = self._get_local_name(str(term.path))
                tname = tname.replace(str(term.path), aname)

            tname = tname[5:-1]

        return tname


    @classmethod
    def _get_node_attr (
        cls,
        term: rdflib.term.Identifier,
        ) -> typing.Dict[str, typing.Any]:
        """
Semiprivate helper function to configure the PyVis node attributes for the given RDF term.

    term:
RDF term

    returns:
PyVis node attributes
        """
        node_attr: typing.Dict[str, typing.Any] = {
            "shape": "ellipse",
        }

        if isinstance(term, rdflib.term.Variable):
            node_attr["style"] = "dashed"
        elif isinstance(term, rdflib.term.BNode):
            node_attr["style"] = "dotted"
        elif isinstance(term, rdflib.term.Literal):
            node_attr["shape"] = "box"

        if "style" in node_attr:
            if node_attr["style"] == "dashed":
                node_attr["shapeProperties"] = { "borderDashes": [5, 5] }
            elif node_attr["style"] == "dotted":
                node_attr["shapeProperties"] = { "borderDashes": [2, 2] }

        return node_attr


    def _render_triples (
        self,
        pyvis_graph: pyvis.network.Network,
        ) -> None:
        """
Semiprivate helper function to render the list of triples extracted from the SPARQL query into PyVis nodes and edges in the given PyVis graph.

    pyvis_graph:
PyVis graph into which the triples get rendered
        """
        for graph_id, triples in enumerate(self.triples):
            for s, p, o in triples:
                # get term labels
                sname = self._get_label(s)
                pname = self._get_label(p)
                oname = self._get_label(o)

                # customize edge attribute
                edge_args: typing.Dict[str, typing.Any] = {
                    "color": self._DEFAULT_COLORS[graph_id + 1],
                    "label": pname,
                    "arrows": { "to": {"enabled": True} },
                }

                if isinstance(p, rdflib.term.Variable):
                    edge_args["style"] = "dashed"
                    edge_args["dashes"] = [5, 5]

                # customize node attribute
                snode_args = self._get_node_attr(s)
                onode_args = self._get_node_attr(o)

                clr: typing.Dict[str, typing.Any] = {
                    "border": self._DEFAULT_COLORS[graph_id + 1],
                    "background": "white",
                }

                snode_args["color"] = clr
                onode_args["color"] = clr

                # add triple
                pyvis_graph.add_node(sname, **snode_args)
                pyvis_graph.add_node(oname, **onode_args)
                pyvis_graph.add_edge(sname, oname, **edge_args)


    def _render_literals (
        self,
        pyvis_graph: pyvis.network.Network,
        ) -> None:
        """
Semiprivate helper function to render the list of literal values (if any) extracted from the SPARQL query.

    pyvis_graph:
PyVis graph into which the triples get rendered
        """
        for var in self.values:
            lname = str(var)
            varname = "?" + lname

            for value in self.values[lname]:
                valname = self._get_label(value)

                edge_args = {
                    "style": "dashed",
                    "dashes": [5, 5],
                    "dir": "none",
                    "arrows": { "to": {"enabled": False} },
                }

                pyvis_graph.add_edge(valname, varname, **edge_args)
                pyvis_graph.nodes[valname]["shape"] = "box"


    def visualize_query (
        self,
        *,
        notebook: bool = False,
        ) -> pyvis.network.Network:
        """
Visualize the SPARQL query as a PyVis network.

        returns:
PyVis graph to be rendered
        """
        pyvis_graph = pyvis.network.Network(notebook=notebook)

        self._render_triples(pyvis_graph)
        self._render_literals(pyvis_graph)

        return pyvis_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import kglab

# create a KnowledgeGraph object
kg = kglab.KnowledgeGraph()

# load RDF from a URL
kg.load_rdf("https://storage.googleapis.com/kglab-tutorial/foaf.rdf", format="xml")

# measure the graph
measure = kglab.Measure()
measure.measure_graph(kg)

print("edges: {}\n".format(measure.get_edge_count()))
print("nodes: {}\n".format(measure.get_node_count()))

# serialize as a string in "Turtle" TTL format
ttl = kg.save_rdf_text()
print("```")
print(ttl[:999])
print("```")



"""
see license https://github.com/DerwenAI/kglab#license-and-copyright
"""
from .kglab import KnowledgeGraph

from .graph import NodeRef, PropertyStore

from .subg import Subgraph, SubgraphMatrix, SubgraphTensor

from .topo import Measure, Simplex0, Simplex1

from .srl import PSLModel

from .esp import ShapeFactory, EvoShape, EvoShapeNode, EvoShapeEdge, \
    Leaderboard, SerializedEvoShape, EvoShapeBoard, EvoShapeDistance

from .pkg_types import RDF_Triple, RDF_Node, SPARQL_Bindings, \
    PathLike, IOPathLike, \
    NodeLike, GraphLike, \
    Census_Item, Census_Dyad_Tally

from .external_import import import_from_neo4j

from .gpviz import GPViz

from .util import get_gpu_count, calc_quantile_bins, stripe_column, root_mean_square

from .version import MIN_PY_VERSION, _versify, _check_version, __version__



# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Mixin definition for `KnowledgeGraph` querying functionalities
"""

## Python standard libraries
import re
import typing

### third-parties libraries
import pandas as pd  # type: ignore
import pyvis  # type: ignore

import rdflib  # type: ignore
import rdflib.plugin  # type: ignore

## kglab - core classes
from ..pkg_types import RDF_Node
from ..gpviz import GPViz
from ..util import get_gpu_count, Mixin
from ..version import _check_version


## pre-constructor set-up
_check_version()

if get_gpu_count() > 0:
    import cudf  # type: ignore  # pylint: disable=E0401


class QueryingMixin(Mixin):
    """
This class implements querying for `KnowledgeGraph`

Core feature areas include:
  * SPARQL querying
  * Cypher querying (future)

Authored by: Paco Nathan
    """
    ######################################################################
    ## SPARQL queries

    def query (
        self,
        sparql: str,
        *,
        bindings: dict = None,
        ) -> typing.Iterable:
        """
Wrapper for [`rdflib.Graph.query()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=query#rdflib.Graph.query) to perform a SPARQL query on the RDF graph.

    sparql:
text for the SPARQL query

    bindings:
initial variable bindings

    yields:
[`rdflib.query.ResultRow`](https://rdflib.readthedocs.io/en/stable/_modules/rdflib/query.html?highlight=ResultRow#) named tuples, to iterate through the query result set
        """
        if not bindings:
            bindings = {}

        for row in self._g.query( # type: ignore
                sparql,
                initBindings=bindings,
            ):
            yield row


    def query_as_df (
        self,
        sparql: str,
        *,
        bindings: dict = None,
        simplify: bool = True,
        pythonify: bool = True,
        ) -> pd.DataFrame:
        """
Wrapper for [`rdflib.Graph.query()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=query#rdflib.Graph.query)
to perform a SPARQL query on the RDF graph.

    sparql:
text for the SPARQL query

    bindings:
initial variable bindings

    simplify:
convert terms in each row of the result set into a readable representation for each term, using N3 format

    pythonify:
convert instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
the query result set represented as a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html); uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled
        """
        if not bindings:
            bindings = {}

        row_iter = self._g.query(sparql, initBindings=bindings) # type: ignore

        if simplify:
            rows_list = [ self.n3fy_row(r.asdict(), pythonify=pythonify) for r in row_iter ]
        else:
            rows_list = [ r.asdict() for r in row_iter ]

        if self.use_gpus:
            df = cudf.DataFrame(rows_list)
        else:
            df = pd.DataFrame(rows_list)

        return df


    def visualize_query (
        self,
        sparql: str,
        *,
        notebook: bool = False,
        ) -> pyvis.network.Network:
        """
Visualize the given SPARQL query as a
[`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html#pyvis.network.Network)

    sparql:
input SPARQL query to be visualized

    notebook:
optional boolean flag, whether to initialize the PyVis graph to render within a notebook; defaults to `False`

    returns:
PyVis network object, to be rendered
        """
        return GPViz(sparql, self._ns).visualize_query(notebook=notebook) # type: ignore


    def n3fy (
        self,
        node: RDF_Node,
        *,
        pythonify: bool = True,
        ) -> typing.Any:
        """
Wrapper for RDFlib [`n3()`](https://rdflib.readthedocs.io/en/stable/utilities.html?highlight=n3#serializing-a-single-term-to-n3)
and [`toPython()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=toPython#rdflib.Variable.toPython)
to serialize a node into a human-readable representation using N3 format.

    node:
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node)

    pythonify:
flag to force instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
text (or Python objects) for the serialized node
        """
        if pythonify and isinstance(node, rdflib.term.Literal):
            serialized = node.toPython()
        else:
            serialized = node.n3(self._g.namespace_manager)  # type: ignore

        return serialized


    def n3fy_row (
        self,
        row_dict: dict,
        *,
        pythonify: bool = True,
        ) -> dict:
        """
Wrapper for RDFlib [`n3()`](https://rdflib.readthedocs.io/en/stable/utilities.html?highlight=n3#serializing-a-single-term-to-n3) and [`toPython()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=toPython#rdflib.Variable.toPython) to serialize one row of a result set from a SPARQL query into a human-readable representation for each term using N3 format.

    row_dict:
one row of a SPARQL query results, as a `dict`

    pythonify:
flag to force instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
a dictionary of serialized row bindings
        """
        bindings = {
            k: self.n3fy(v, pythonify=pythonify)
            for k, v in row_dict.items()
        }

        return bindings


    @classmethod
    def unbind_sparql (
        cls,
        sparql: str,
        bindings: dict,
        *,
        preamble: str = "",
        ) -> str:
        """
Substitute the _binding variables_ into the text of a SPARQL query,
to obviate the need for binding variables specified separately.
This can be helpful for debugging, or for some query engines that
may not have full SPARQL support yet.

    sparql:
text for the SPARQL query

    bindings:
variable bindings

    returns:
a string of the expanded SPARQL query
        """
        sparql_meta, sparql_body = re.split(r"\s*WHERE\s*\{", sparql, maxsplit=1)

        for var in sorted(bindings.keys(), key=lambda x: len(x), reverse=True):  # pylint: disable=W0108
            pattern = re.compile(r"(\?" + var + r")(\W)")  # pylint: disable=W1401
            bind_val = "<" + str(bindings[var]) + ">\\2"
            sparql_body = re.sub(pattern, bind_val, sparql_body)

        result = "".join([preamble, sparql_meta, " WHERE {", sparql_body]).strip()
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Shared type definitions.
"""

import pathlib
import typing

import pandas as pd  # type: ignore  # pylint: disable=E0401
import rdflib  # type: ignore  # pylint: disable=E0401
import urlpath  # type: ignore  # pylint: disable=E0401


PathLike = typing.Union[ str, pathlib.Path, urlpath.URL ]
IOPathLike = typing.Union[ PathLike, typing.IO ]

RDF_Node = typing.Union[ rdflib.term.Node, rdflib.term.URIRef, rdflib.term.Literal, rdflib.term.BNode ]
RDF_Triple = typing.Tuple[ RDF_Node, RDF_Node, RDF_Node ]
NodeLike = typing.Union[ typing.Optional[str], RDF_Node ]

ConjunctiveLike = typing.Union[ rdflib.ConjunctiveGraph, rdflib.Dataset ]
GraphLike = typing.Union[ ConjunctiveLike, rdflib.Graph ]

SPARQL_Bindings = typing.Tuple[ str, dict ]

Census_Item = typing.Union[ str, RDF_Node ]
Census_Dyad_Tally = typing.Tuple[ pd.DataFrame, dict ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Support for use of statistical relational learning.
"""

import pathlib
import typing

from icecream import ic  # type: ignore  # pylint: disable=E0401
import pandas as pd  # type: ignore  # pylint: disable=E0401
import pslpython.model  # type: ignore  # pylint: disable=E0401
import pslpython.partition  # type: ignore  # pylint: disable=E0401
import pslpython.predicate  # type: ignore  # pylint: disable=E0401
import pslpython.rule  # type: ignore  # pylint: disable=E0401


class PSLModel:
    """
Class representing a
[*probabilistic soft logic*](../glossary/#probabilistic-soft-logic)
(PSL) model.

For PSL-specific terminology used here, see <https://psl.linqs.org/wiki/master/Glossary.html>

Note: You need to have a Java JDK installed to run PSL.
    """

    _PSL_OPTIONS: dict = {
        "log4j.threshold": "INFO",
    }

    _PARTITIONS: dict = {
        "observations": pslpython.partition.Partition.OBSERVATIONS,
        "targets": pslpython.partition.Partition.TARGETS,
        "truth": pslpython.partition.Partition.TRUTH,
    }


    def __init__ (
        self,
        *,
        name: str = None,
        ) -> None:
        """
Wrapper for constructing a [`pslpython.model.Model`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/model.py).

    name:
optional name of the PSL model; if not supplied, PSL generates a random name
        """
        self.model = pslpython.model.Model(name)
        self.results: dict = {}


    def clear_model (
        self
        ) -> "PSLModel":
        """
Clear any pre-existing data from each of the predicates, to initialize the model.

    returns:
this PSL model – use for method chaining
        """
        for predicate in self.model.get_predicates().values():
            predicate.clear_data()

        return self


    def add_predicate (
        self,
        raw_name: str,
        *,
        size: int = None,
        closed: bool = False,
        arg_types: typing.List = None,
        ) -> "PSLModel":
        """
Add a [`pslpython.predicate.Predicate`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/predicate.py) to this model.
Enough details must be supplied for PSL to infer the number and types of each predicate's arguments.

    raw_name:
name of the predicate; must be unique among all of the predicates

    size:
optional, the number of arguments for this predicate

    closed:
indicates that this predicate is fully observed, i.e., all substitutions of this predicate have known values and will behave as evidence for inference; otherwise, if `False` then infer some values of this predicate; defaults to `False`

    arg_types:
optional, a list of types for the arguments for this predicate; all arguments will default to string

    returns:
this PSL model – use for method chaining
        """
        predicate = pslpython.predicate.Predicate(
            raw_name,
            closed=closed,
            size=size,
            arg_types=arg_types,
            )

        self.model.add_predicate(predicate)
        return self


    def add_rule (
        self,
        rule_string: str,
        *,
        weighted: bool = None,
        weight: float = None,
        squared: bool = None,
        ) -> "PSLModel":
        """
Add a [`pslpython.rule.Rule`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/rule.py) to this model.

  * a weighted rule can change its weight or squared status
  * a weighted rule cannot convert into an unweighted rule nor visa-versa
  * unweighted rules are [*constraints*](https://psl.linqs.org/wiki/master/Constraints.html)

For more details, see <https://psl.linqs.org/wiki/master/Rule-Specification.html>

    rule_string:
text representation for specifying the rule

    weighted:
indicates that this rule is weighted

    weight:
weight of this rule

    squared:
indicates that this rule's potential is squared

    returns:
this PSL model – use for method chaining
        """
        if weight:
            weighted = True

        rule = pslpython.rule.Rule(
            rule_string=rule_string,
            weighted=weighted,
            weight=weight,
            squared=squared,
        )

        self.model.add_rule(rule)
        return self


    @classmethod
    def _raise_model_error (
        cls,
        obj: str,  # pylint: disable=W0613
        msg: str,  # pylint: disable=W0613
        ) -> None:
        """
Semiprivate helper function to format and raise a `ModelError` exception.

    obj:
name of the object to be reported

    msg:
the exception message to use
        """
        error = "{msg}: {obj}"
        raise pslpython.model.ModelError(error)


    def _get_predicate (
        self,
        predicate_name: str,
        ) -> pslpython.predicate.Predicate:
        """
Semiprivate accessor method to lookup a predicate, with error checking.

    predicate_name:
name of the specific predicate; name normalization will be handled internally; raises `ModelError` if the predicate name is not found

    returns:
predicate object
        """
        try:
            predicate = self.model.get_predicate(predicate_name)

            if not predicate:
                self._raise_model_error(predicate_name, "Unknown predicate")
        except:  # pylint: disable=W0702 # lgtm[py/catch-base-exception]
            self._raise_model_error(predicate_name, "Unknown predicate")

        return predicate


    def _get_partition (
        self,
        partition: str,
        ) -> pslpython.partition.Partition:
        """
Semiprivate accessor method to lookup a partition, with error checking.

    partition:
label for the [`pslpython.partition.Partition`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/partition.py) into which the `data` gets added; must be among `[ "observations", "targets", "truth" ]`; see <https://psl.linqs.org/wiki/master/Data-Storage-in-PSL.html>

    returns:
partition object
        """
        try:
            partition_obj = self._PARTITIONS[partition.lower()]

            if not partition_obj:
                self._raise_model_error(partition, "Unknown partition")
        except:  # pylint: disable=W0702 # lgtm[py/catch-base-exception]
            self._raise_model_error(partition, "Unknown partition")

        return partition_obj


    def add_data_row (
        self,
        predicate_name: str,
        args: list,
        *,
        partition: str = "observations",
        truth_value: float = 1.0,
        verbose: bool = False,
        ) -> "PSLModel":
        """
Add a single record to a specified predicate, within a specified partition.

    predicate_name:
name of the specific predicate; name normalization will be handled internally; raises `ModelError` if the predicate name is not found

    args:
arguments for the record being added, as a list

    partition:
label for the [`pslpython.partition.Partition`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/partition.py) into which the `data` gets added; must be among `[ "observations", "targets", "truth" ]`; defaults to `"observations"`; see <https://psl.linqs.org/wiki/master/Data-Storage-in-PSL.html>

    truth_value:
optional truth value of the record being added

    verbose:
flag for verbose trace of each added record

    returns:
this PSL model – use for method chaining
        """
        if verbose:
            ic(predicate_name, partition, args)

        predicate = self._get_predicate(predicate_name)

        predicate.add_data_row(
            self._get_partition(partition),
            args=args,
            truth_value=truth_value,
        )

        return self


    def trace_predicate (
        self,
        predicate_name: str,
        *,
        partition: str = "observations",
        path: pathlib.Path = None,
        ) -> pd.DataFrame:
        """
Construct a trace of the data in a specified predicate, within a specified partition, formatted as a dataframe.
Use a consistent column naming and sort order, so that these values can be used later in testing.
Optionally write out this out to a TSV file.

    predicate_name:
name of the specific predicate; name normalization will be handled internally; raises `ModelError` if the predicate name is not found

    partition:
label for the [`pslpython.partition.Partition`](https://github.com/linqs/psl/blob/master/psl-python/pslpython/partition.py) into which the `data` gets added; must be among `[ "observations", "targets", "truth" ]`; defaults to `"observations"`; see <https://psl.linqs.org/wiki/master/Data-Storage-in-PSL.html>

    path:
optional output path for the TSV file; defaults to `None`

    returns:
dataframe representing the traced partition data
        """
        predicate = self._get_predicate(predicate_name)
        partition_obj = self._get_partition(partition)

        df = predicate._data[partition_obj].copy(deep=True)  # pylint: disable=W0212
        df.columns = [ "P1", "P2", "value" ]
        df = df.sort_values(by=[ "P1", "P2" ]).reset_index(drop=True)

        if path:
            df.to_csv(path, sep="\t", index=False)

        return df


    @classmethod
    def compare_predicate (
        cls,
        df: pd.DataFrame,
        trace_path: pathlib.Path,
        ) -> pd.DataFrame:
        """
Compare the values of a predict with its expected values which get loaded from a file.
This will print any expected (missing) or error (mismatched) rows.

    df:
dataframe from `trace_predicate`

    trace_path:
path to a TSV file of expected values, saved from the trace of a baseline run

    returns:
dataframe loaded from the expected values
        """
        df_known = pd.read_csv(trace_path, sep="\t")
        df_known[[ "P1", "P2" ]] = df_known[[ "P1", "P2" ]].astype(int)
        df_known = df_known.sort_values(by=[ "P1", "P2" ]).reset_index(drop=True)

        for _, row in df_known.iterrows():
            p1, p2, value = row
            loc = df.loc[(df["P1"] == p1) & (df["P2"] == p2)]

            if len(loc) > 0:
                loc_val = float(loc["value"])

                if value != loc_val:
                    print(" expected", loc)
            else:
                print(" error", int(p1), int(p2), value)

        return df_known


    def infer (
        self,
        *,
        method: str = "",
        cli_options: list = None,
        psl_config: dict = None,
        jvm_options: list = None,
        ) -> None:
        """
Run inference on this model, storing the inferred results in an internal dataframe.

    method:
the inference method to use

    cli_options:
additional options to pass to PSL, based on its CLI options; see <https://psl.linqs.org/wiki/master/Configuration.html>

    psl_config:
configuration options passed directly to the PSL core code; see <https://psl.linqs.org/wiki/master/Configuration-Options.html>

    jvm_options:
options passed to the JVM running the PSL Java library; most commonly `"-Xmx"` and `"-Xms"`
        """
        if not cli_options:
            cli_options = []

        if not psl_config:
            psl_config = self._PSL_OPTIONS

        if not jvm_options:
            jvm_options = []

        self.results = self.model.infer(
            method=method,
            additional_cli_optons=cli_options,
            psl_config=psl_config,
            jvm_options=jvm_options,
        )


    def get_results (
        self,
        predicate_name: str,
        ) -> pd.DataFrame:
        """
Accessor for the inferred results for a specified predicate.

    predicate_name:
name of the specific predicate; name normalization will be handled internally; raises `ModelError` if the predicate name is not found

    returns:
inferred values as a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html), with columns names for each argument plus the `"truth"` value
        """
        predicate = self._get_predicate(predicate_name)
        df = self.results[predicate].copy(deep=True)
        df.insert(0, "predicate", predicate.name())

        return df



# type: ignore

import importlib.util
import pathlib
import setuptools
import typing


KEYWORDS = [
    "controlled vocabulary",
    "cugraph",
    "deep learning",
    "embedding",
    "gpu",
    "graph algorithms",
    "igraph",
    "inference",
    "interactive visualization",
    "json-ld",
    "knowledge graph",
    "managing namespaces",
    "morph-kgc",
    "n3",
    "networkx",
    "owl",
    "pandas",
    "parquet",
    "probabilistic soft logic",
    "psl",
    "pyvis",
    "rapids",
    "rdf",
    "rml",
    "roam research",
    "serialization",
    "shacl",
    "skos",
    "sparql",
    "statistical relational learning",
    "topology",
    "turtle",
    "validation",
    ]


def parse_requirements_file (
    filename: str,
    ) -> typing.List[ str ]:
    """read and parse a Python `requirements.txt` file, returning as a list of strings"""
    with pathlib.Path(filename).open() as f:
        results: list = [
            l.strip().replace(" ", "").split("#")[0]
            for l in f.readlines()
        ]

        return results


if __name__ == "__main__":
    spec = importlib.util.spec_from_file_location("kglab.version", "kglab/version.py")
    kglab_version = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(kglab_version)
    kglab_version._check_version()  # pylint: disable=W0212

    base_packages = parse_requirements_file("requirements.txt")
    docs_packages = parse_requirements_file("requirements-dev.txt")
    tut_packages = parse_requirements_file("requirements-tut.txt")

    setuptools.setup(
        name = "kglab",
        version = kglab_version.__version__,

        license = "MIT",
        description = "A simple abstraction layer in Python for building knowledge graphs",
        long_description = pathlib.Path("README.md").read_text(),
        long_description_content_type = "text/markdown",

        python_requires = ">=" + kglab_version._versify(kglab_version.MIN_PY_VERSION),  # pylint: disable=W0212
        zip_safe = False,

        packages = setuptools.find_packages(
            exclude = [
                "dat",
                "docs",
                "examples",
                "scripts",
                "tests",
            ]),

        install_requires = base_packages,
        extras_require = {
            "base": base_packages,
            "docs": docs_packages,
            "tutorial": tut_packages,
        },

        author = "Paco Nathan",
        author_email = "paco@derwen.ai",

        keywords = ", ".join(KEYWORDS),
        classifiers = [
            "Programming Language :: Python",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Operating System :: OS Independent",
            "License :: OSI Approved :: MIT License",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Information Technology",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Human Machine Interfaces",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Text Processing :: Indexing",
            ],

        url = "https://derwen.ai/docs/kgl/",
        project_urls = {
            "DOI": "https://doi.org/10.5281/zenodo.6360664",
            "Discussions": "https://www.linkedin.com/groups/6725785/",
            "DockerHub": "https://hub.docker.com/r/derwenai/kglab",
            "Tutorial": "https://derwen.ai/docs/kgl/tutorial/",
            "Issues": "https://github.com/DerwenAI/kglab/issues",
            "Source": "https://github.com/DerwenAI/kglab",
        },

        entry_points = {
            "rdf.plugins.store": [
                "kglab = kglab:PropertyStore",
            ],
        },
    )



"""
Main class definitions for `kglab`

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

## Python standard libraries
import traceback
import typing

### third-parties libraries
from icecream import ic  # type: ignore
import dateutil.parser as dup
import pandas as pd  # type: ignore

import rdflib  # type: ignore
import rdflib.plugin  # type: ignore

## kglab - core classes
from .pkg_types import GraphLike, RDF_Node
from .util import get_gpu_count
from .version import _check_version
from .query.sparql import SparqlQueryable
from .query.mixin import QueryingMixin
from .serde import SerdeMixin
from .standards import ShaclOwlRdfSkosMixin


## pre-constructor set-up
_check_version()

if get_gpu_count() > 0:
    import cudf  # type: ignore


class KnowledgeGraph(QueryingMixin, SerdeMixin, ShaclOwlRdfSkosMixin):
    """
This is the primary class used to represent RDF graphs, on which the other classes are dependent.
See <https://derwen.ai/docs/kgl/concepts/#knowledge-graph>

Core feature areas include:

  * namespace management: ontology, controlled vocabularies
  * graph construction
  * serialization-deserilization (see `serde` module)
  * SPARQL querying (see `query.mixin` module)
  * SHACL validation (see `standards` module)
  * inference based on OWL-RL, RDFS, SKOS (see `standards` module)
    """

    _DEFAULT_NAMESPACES: dict = {
        "dct":    "http://purl.org/dc/terms/",
        "owl":    "http://www.w3.org/2002/07/owl#",
        "prov":   "http://www.w3.org/ns/prov#",
        "rdf":    "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs":   "http://www.w3.org/2000/01/rdf-schema#",
        "schema": "http://schema.org/",
        "sh":     "http://www.w3.org/ns/shacl#",
        "skos":   "http://www.w3.org/2004/02/skos/core#",
        "xsd":    "http://www.w3.org/2001/XMLSchema#",
    }


    def __init__ (
        self,
        *,
        name: str = "generic",
        base_uri: typing.Optional[str] = None,
        language: str = "en",
        store: typing.Optional[str] = None,
        use_gpus: bool = True,
        import_graph: typing.Optional[GraphLike] = None,
        namespaces: typing.Optional[dict] = None,
        ) -> None:
        """
Constructor for a `KnowledgeGraph` object.

    name:
optional, internal name for this graph

    base_uri:
the default [*base URI*](https://tools.ietf.org/html/rfc3986#section-5.1) for this RDF graph

    language:
the default [*language tag*](https://www.w3.org/TR/rdf11-concepts/#dfn-language-tag), e.g., used for [*language indexing*](https://www.w3.org/TR/json-ld11/#language-indexing)

    store:
optionally, string representing an `rdflib.Store` plugin to use.

    use_gpus:
optionally, use the NVidia GPU devices with [RAPIDS](https://rapids.ai/) if these libraries have been installed and the devices are available; defaults to `True`

    import_graph:
optionally, another existing RDF graph to be used as a starting point

    namespaces:
a dictionary of [*namespace*](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=namespace#rdflib.Namespace) (dict values) and their corresponding *prefix* strings (dict keys) to add as *controlled vocabularies* which are available for use in the RDF graph, binding each prefix to the given namespace
        """
        self.name = name
        self.base_uri = base_uri
        self.language = language
        self.store = store

        # use NVidia GPU devices if available and the libraries
        # have been installed and the flag is not disabled
        if use_gpus is not None and get_gpu_count() > 0:
            self.use_gpus = True
        else:
            self.use_gpus = False

        # import relations from another RDF graph, or start from blank
        if import_graph is not None:
            self._g = import_graph
        else:
            self._g = self.build_blank_graph() # pylint: disable=E1101

        # initialize the namespaces
        self._ns: dict = {}

        for prefix, iri in self._DEFAULT_NAMESPACES.items():
            self.add_ns(prefix, iri) # pylint: disable=E1101

        if namespaces:
            for prefix, iri in namespaces.items():
                self.add_ns(prefix, iri) # pylint: disable=E1101

        # backwards compatibility for class refactoring
        self.sparql = SparqlQueryable(self)


    def build_blank_graph (
        self,
        ) -> rdflib.Graph:
        """
Build a new `rdflib.Graph` object, based on storage plugin configuration.
        """
        if self.store is not None:
            g = rdflib.Graph(store=self.store)
        else:
            g = rdflib.Graph()

        return g


    def rdf_graph (
        self
        ) -> rdflib.Graph:
        """
Accessor for the RDF graph.

    returns:
the [`rdflib.Graph`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=graph#graph) object
        """
        return self._g


    ######################################################################
    ## namespace management and graph building
    ##
    ## Using and building ontologies: To attribute characteristics and
    ## relationships to well-understood entities, we want to research
    ## the implementation of existing top-level ontologies. In such an
    ## implementation, however, we would not want to prevent mid-level
    ## and domain specific ontologies from being developed organically
    ## for classification with greater precision and nuance. We want
    ## to explore ways to perform entity recognition and
    ## entity-resolution probabilistically rather than by strict
    ## rulesets.

    def add_ns (
        self,
        prefix: str,
        iri: str,
        override: bool = True,
        replace: bool = False,
        ) -> None:
        """
Adds another [*namespace*](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=namespace#rdflib.Namespace) among the *controlled vocabularies* available to use in the RDF graph, binding the `prefix` to the given namespace.

Since the RDFlib [`NamespaceManager`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=namespace#rdflib.namespace.NamespaceManager) automagically converts all input bindings into [`URIRef`](https://www.w3.org/TR/rdf-concepts/#section-Graph-URIref) instead, we'll keep references to the namespaces – for later use.

    prefix:
a [namespace prefix](https://www.w3.org/TR/rdf11-concepts/#dfn-namespace-prefix); it's recommended to confirm prefix usage (based on convention) by searching on <http://prefix.cc/>

    iri:
URL to use for constructing the [namespace IRI](https://www.w3.org/TR/rdf11-concepts/#dfn-namespace-iri)

    override:
rebind, even if the given namespace is already bound with another prefix

    replace:
replace any existing prefix with the new namespace
        """
        if override and iri in self._ns.values():
            rev_ns = {
                str(v): k
                for k, v in self._ns.items()
            }

            bogus_prefix = rev_ns[iri]
            del self._ns[bogus_prefix]

        if replace or prefix not in self._ns:
            self._ns[prefix] = rdflib.Namespace(iri)

        self._g.namespace_manager.bind( # type: ignore
            prefix,
            self._ns[prefix],
            override=override,
            replace=replace,
        )


    def get_ns (
        self,
        prefix: str,
        ) -> rdflib.Namespace:
        """
Lookup a [*namespace*](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=namespace#rdflib.Namespace) among the *controlled vocabularies* available to use in the RDF graph.

    prefix:
a [namespace prefix](https://www.w3.org/TR/rdf11-concepts/#dfn-namespace-prefix)

    returns:
the RDFlib [`Namespace`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=namespace#rdflib.Namespace) for the *controlled vocabulary* referenced by `prefix`
        """
        return self._ns[prefix]


    def get_ns_dict (
        self
        ) -> dict:
        """
Generate a dictionary of the *namespaces* used in this RDF graph.

    returns:
a `dict` describing the namespaces in this RDF graph
        """
        ns_dict = {
            prefix: str(ns)
            for prefix, ns in self._ns.items()
        }

        nm = self._g.namespace_manager  # type: ignore

        for prefix, uri in nm.namespaces():
            ns_dict[prefix] = str(uri)

        return ns_dict


    def describe_ns (
        self
        ) -> pd.DataFrame:
        """
Describe the *namespaces* used in this RDF graph.

    returns:
a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html) describing the namespaces in this RDF graph; uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled
        """
        col_names: typing.List[str] = [ "prefix", "namespace" ]

        rows_list: typing.List[dict] = [
            {
                col_names[0]: prefix,
                col_names[1]: str(ns),
            }
            for prefix, ns in self.get_ns_dict().items()
        ]

        if self.use_gpus:
            df = cudf.DataFrame(rows_list, columns=col_names)
        else:
            df = pd.DataFrame(rows_list, columns=col_names)

        return df


    def get_context (
        self
        ) -> dict:
        """
Generates a [*JSON-LD context*](https://www.w3.org/TR/json-ld11/#the-context) used for
serializing the RDF graph as [JSON-LD](https://json-ld.org/).

    returns:
context needed for JSON-LD serialization
        """
        context: dict = self.get_ns_dict()
        context["@language"] = self.language

        if self.base_uri:
            context["@vocab"] = self.base_uri

        return context


    def encode_date (
        self,
        dt: str,
        tzinfos: dict,
        ) -> rdflib.Literal:
        """
Helper method to ensure that an input `datetime` value has a timezone that can be interpreted by [`rdflib.XSD.dateTime`](https://www.w3.org/TR/xmlschema-2/#dateTime).

    dt:
input datetime as a string

    tzinfos:
timezones as a dict, used by
[`dateutil.parser.parse()`](https://dateutil.readthedocs.io/en/stable/parser.html#dateutil.parser.parse) as additional time zone names or aliases which may be present in the input `datetime` string

    returns:
[`rdflib.Literal`](https://rdflib.readthedocs.io/en/stable/rdf_terms.html#literals) formatted as an XML Schema 2 `dateTime` value
        """
        date_tz = dup.parse(dt, tzinfos=tzinfos)
        return rdflib.Literal(date_tz, datatype=self.get_ns("xsd").dateTime) # pylint: disable=E1101


    def add (
        self,
        s: RDF_Node,
        p: RDF_Node,
        o: RDF_Node,
        ) -> None:
        """
Wrapper for [`rdflib.Graph.add()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=add#rdflib.Graph.add) to add a relation *(subject, predicate, object)* to the RDF graph, if it doesn't already exist.
Uses the RDF Graph as its context.

To prepare for upcoming **kglab** features, **this is the preferred method for adding relations to an RDF graph.**

    s:
*subject* node;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node); otherwise throws a `TypeError` exception

    p:
*predicate* relation;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node); otherwise throws a `TypeError` exception

    o:
*object* node;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node) or [`rdflib.term.Terminal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Literal); otherwise throws a `TypeError` exception
        """
        try:
            self._g.add((s, p, o,))  # type: ignore
        except AssertionError as e:
            traceback.print_exc()
            ic(s)
            ic(p)
            ic(o)
            raise TypeError(str(e))


    def remove (
        self,
        s: RDF_Node,
        p: RDF_Node,
        o: RDF_Node,
        ) -> None:
        """
Wrapper for [`rdflib.Graph.remove()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=add#rdflib.Graph.remove) to remove a relation *(subject, predicate, object)* from the RDF graph, if it exist.
Uses the RDF Graph as its context.

To prepare for upcoming **kglab** features, **this is the preferred method for removing relations from an RDF graph.**

    s:
*subject* node;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node); otherwise throws a `TypeError` exception

    p:
*predicate* relation;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node); otherwise throws a `TypeError` exception

    o:
*object* node;
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node) or [`rdflib.term.Terminal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Literal); otherwise throws a `TypeError` exception
        """
        try:
            self._g.remove((s, p, o,)) # type: ignore
        except AssertionError as e:
            traceback.print_exc()
            ic(s)
            ic(p)
            ic(o)
            raise TypeError(str(e))


    def graph_factory(self, name, graph):
        """
Utility function to generate graphs from mixins

        name:
name of the graph

        graph:
initial graph
        """
        return KnowledgeGraph(
            name=name,
            namespaces=self.get_ns_dict(),
            import_graph=graph,
        )
